# -*- coding: utf-8 -*-
"""
Created on Tue Mar 17 14:41:31 2015

Read gait trials.

@author: Jussi (jnu@iki.fi)
"""


from __future__ import division
from builtins import range
from builtins import object
from collections import defaultdict
import numpy as np
import re
import os.path as op
import glob
import logging

from . import read_data
from . import nexus
from . import utils
from . import eclipse
from . import models
from . import videos
from .emg import EMG
from .config import cfg
from .envutils import GaitDataError


logger = logging.getLogger(__name__)


def nexus_trial(from_c3d=False):
    """Return Trial instance from currently open Nexus trial."""
    vicon = nexus.viconnexus()
    if from_c3d:
        trname = vicon.GetTrialName()
        c3dfile = op.join(*trname) + '.c3d'
        if op.isfile(c3dfile):
            return Trial(c3dfile)
        else:
            logger.info('no c3d file for currently loaded trial, loading '
                        'directly from Nexus')
            return Trial(nexus.viconnexus())
    else:
        return Trial(nexus.viconnexus())


def _nexus_crop_events_before_forceplate():
    """Delete events preceding first forceplate strike."""
    vicon = nexus.viconnexus()
    tr = Trial(vicon)
    fp_cycles = [cyc for cyc in tr.cycles if cyc.on_forceplate]
    cycs_sorted = sorted(fp_cycles, key=lambda cyc: cyc.start)
    if not cycs_sorted:
        return
    cyc1 = cycs_sorted[0]
    context = {'L': 'Left', 'R': 'Right'}[cyc1.context]
    context_other = 'Left' if context == 'Right' else 'Right'
    strike_ctxt = vicon.GetEvents(tr.name, context, "Foot Strike")[0]
    strike_ctxt = [f for f in strike_ctxt if f >= cyc1.start]
    toeoff_ctxt = vicon.GetEvents(tr.name, context, "Foot Off")[0]
    toeoff_ctxt = [f for f in toeoff_ctxt if f >= cyc1.start]
    strike_other = vicon.GetEvents(tr.name, context_other, "Foot Strike")[0]
    toeoff_other = vicon.GetEvents(tr.name, context_other, "Foot Off")[0]
    vicon.ClearAllEvents()
    for ev in strike_ctxt:
        vicon.CreateAnEvent(tr.name, context, 'Foot Strike', ev, 0)
    for ev in strike_other:
        vicon.CreateAnEvent(tr.name, context_other, 'Foot Strike', ev, 0)
    for ev in toeoff_ctxt:
        vicon.CreateAnEvent(tr.name, context, 'Foot Off', ev, 0)
    for ev in toeoff_other:
        vicon.CreateAnEvent(tr.name, context_other, 'Foot Off', ev, 0)


class Noncycle(object):
    """Used in place of Gaitcycle when requesting unnormalized data."""
    def __init__(self, context, trial=None):
        self.context = context
        self.trial = trial
        self.name = 'unnormalized (%s)' % context
        self.toeoffn = None
        self.on_forceplate = False
        self.start = None
        self.end = None


class Gaitcycle(object):
    """ Gait cycle class.

        Parameters
        ----------
        start : int
            Starting frame for the cycle.
        start : int
            Ending frame for the cycle.
        toeoff : int
            Frame where toeoff occurs.
        context : str
            Cycle context: R or L for right and left, respectively.
        on_forceplate : bool
            Whether cycle starts on forceplate contact.
        plate_idx : int
            Index of forceplate.
        smp_per_frame : float
            Analog samples per frame.
        trial : instance of Trial
            The trial instance owning this cycle. Does not need to be set.
        name : str
            Name for the cycle. Can be set freely.
        index : int
            Cycle index.
    """

    def __init__(self, start, end, toeoff, context, on_forceplate, plate_idx,
                 smp_per_frame, trial=None, name=None, index=None):
        self.len = end - start
        # convert frame indices to 0-based
        self.start = start
        self.end = end
        self.toeoff = toeoff
        # which foot begins and ends the cycle
        self.context = context
        # whether cycle begins with forceplate strike
        self.on_forceplate = on_forceplate
        # start and end on the analog samples axis; round to whole samples
        self.start_smp = int(round(self.start * smp_per_frame))
        self.end_smp = int(round(self.end * smp_per_frame))
        self.len_smp = self.end_smp - self.start_smp
        # normalized x-axis (% of gait cycle) of same length as cycle
        self.t = np.linspace(0, 100, self.len)
        # same for analog variables
        self.tn_analog = np.linspace(0, 100, self.len_smp)
        # normalized x-axis of 0,1,2..100%
        self.tn = np.linspace(0, 100, 101)
        # normalize toe-off event to the cycle
        self.toeoffn = round(100*((self.toeoff - self.start) / self.len))
        self.trial = trial
        self.plate_idx = plate_idx
        self.index = index
        self.name = name

    def __repr__(self):
        s = '<Gaitcycle |'
        s += ' start: %d,' % self.start
        s += ' end: %d,' % self.end
        s += ' context: %s,' % self.context
        s += ' on forceplate,' if self.on_forceplate else ' not on forceplate,'
        s += ' toeoff: %d' % self.toeoff
        s += '>'
        return s

    def normalize(self, var):
        """Normalize (columns of) frames-based variable var to the cycle.

        Parameters
        ----------
        var : array
            NxM array of frame-based data to normalize.
        """
        # convert 1D arrays to 2D
        if len(var.shape) == 1:
            var = var[:, np.newaxis]
        ncols = var.shape[1]
        idata = np.array([np.interp(self.tn,
                                    self.t, var[self.start:self.end, k])
                          for k in range(ncols)]).T
        return self.tn, np.squeeze(idata)

    def crop_analog(self, var):
        """Crop analog variable (EMG, forceplate, etc. ) to the gait cycle.

        Parameters
        ----------
        var : array
            NxM array of analog data to normalize.
        """
        return self.tn_analog, var[self.start_smp:self.end_smp]


class Trial(object):
    """Gait trial class.

        Parameters
        ----------
        source : str | instance of ViconNexus
            Source to read data from. Can be a c3d filename or a ViconNexus
            connection.
    """

    def __repr__(self):
        s = '<Trial |'
        s += ' trial: %s' % self.trialname
        s += ', data source: %s' % self.source
        s += ', subject: %s' % self.name
        s += ', gait cycles: %s' % self.ncycles
        s += '>'
        return s

    def __init__(self, source):
        logger.debug('new trial instance from %s' % source)
        self.source = source
        meta = read_data.get_metadata(source)
        # insert metadata dict directly as instance attributes
        self.__dict__.update(meta)

        # sort events and make them 0-based so that indexing matches frame data
        self.lstrikes = [e - self.offset for e in sorted(self.lstrikes)]
        self.rstrikes = [e - self.offset for e in sorted(self.rstrikes)]
        self.ltoeoffs = [e - self.offset for e in sorted(self.ltoeoffs)]
        self.rtoeoffs = [e - self.offset for e in sorted(self.rtoeoffs)]

        self.sessiondir = op.split(self.sessionpath)[-1]

        enfpath = op.join(self.sessionpath, '%s.Trial.enf' % self.trialname)
        # also look for alternative (older style?) enf name
        if not op.isfile(enfpath):
            trialn_re = re.search('\.*(\d*)$', self.trialname)
            trialn = trialn_re.group(1)
            if trialn:
                trialname_ = '%s.Trial%s.enf' % (self.trialname, trialn)
                enfpath = op.join(self.sessionpath, trialname_)
        self.enfpath = enfpath
        if op.isfile(self.enfpath):
            logger.debug('reading Eclipse info from %s' % self.enfpath)
            edata = eclipse.get_eclipse_keys(self.enfpath)
            # for convenience, eclipse_data returns '' for nonexistent keys
            self.eclipse_data = defaultdict(lambda: '', edata)
        else:
            logger.debug('no .enf file found')
            self.eclipse_data = defaultdict(lambda: '', {})
        self.is_static = self.eclipse_data['TYPE'].upper() == 'STATIC'
        # data are lazily read
        self.emg = EMG(self.source)
        self._forceplate_data = None
        self._marker_data = None
        if not self.is_static:
            self.fp_events = self._get_fp_events()
        else:
            self.fp_events = utils.empty_fp_events()

        self._models_data = dict()
        self.stddev_data = None  # AvgTrial only
        # whether to normalize data
        self._normalize = None
        # frames 0...length
        self.t = np.arange(self.length)
        # analog frames 0...length
        self.t_analog = np.arange(self.length * self.samplesperframe)
        # normalized x-axis of 0, 1, 2 .. 100%
        self.tn = np.linspace(0, 100, 101)
        self.samplesperframe = self.analograte/self.framerate
        self.cycles = list() if self.is_static else list(self._scan_cycles())
        self.ncycles = len(self.cycles)

    @property
    def videos(self):
        """Get all trial videos"""
        trialbase = op.join(self.sessionpath, self.trialname)
        return videos.get_trial_videos(trialbase)

    @property
    def eclipse_tag(self):
        """Return the first matching Eclipse tag for this trial."""
        for tag in cfg.eclipse.tags:
            if any([tag in self.eclipse_data[fld] for fld in
                    cfg.eclipse.tag_keys]):
                return tag
        return None

    @property
    def name_with_description(self):
        """Return string consisting of trial name and some Eclipse info."""
        # FIXME: Eclipse keys hardcoded
        return '%s (%s, %s)' % (self.trialname,
                                self.eclipse_data['DESCRIPTION'],
                                self.eclipse_data['NOTES'])

    def _normalized_frame_data(self, data):
        """Return time axis and cycle normalized frame data"""
        if self._normalize is not None:
            t, data = self._normalize.normalize(data)
        else:
            t = self.t
        return t, data

    def _normalized_analog_data(self, data):
        """Return time axis and cycle normalized (cropped) analog data"""
        if self._normalize is not None:
            t, data = self._normalize.crop_analog(data)
        else:
            t = self.t_analog
        return t, data

    @property
    def full_marker_data(self):
        """Return the full marker data dict."""
        if not self._marker_data:
            self._marker_data = read_data.get_marker_data(self.source,
                                                          self.markers)
        return self._marker_data

    def get_model_data(self, var):
        """Return trial data for a model variable.

        Parameters
        ----------
        var : string
            The model variable name.
        """
        data = self._get_modelvar(var)
        return self._normalized_frame_data(data)

    def get_emg_data(self, ch):
        """Return trial data for an EMG variable.

        Parameters
        ----------
        ch : string
            The EMG channel name.
        """
        data = self.emg[ch]
        return self._normalized_analog_data(data)

    def get_marker_data(self, marker):
        """Return trial data for a given marker.

        Parameters
        ----------
        marker : string
            The marker name.
        """
        if marker not in self.markers:
            raise GaitDataError('No such marker')
        data = self.full_marker_data[marker]
        return self._normalized_frame_data(data)

    def get_forceplate_data(self, nplate, kind='force'):
        """Return trial data for a forceplate.

        Parameters
        ----------
        nplate : int
            The forceplate index. Plates are numbered starting from 0, i.e.
            nplate=0 corresponds to Nexus plate 1.
        kind : str
            The type of data to return. Can be 'force', 'moment', or 'cop'
            (center of pressure).
        """
        if not self._forceplate_data:
            self._forceplate_data = read_data.get_forceplate_data(self.source)
        if nplate < 0 or nplate >= len(self._forceplate_data):
            raise GaitDataError('Invalid plate index %d' % nplate)
        if kind == 'force':
            data = self._forceplate_data[nplate]['F']
        elif kind == 'moment':
            data = self._forceplate_data[nplate]['M']
        elif kind == 'cop':
            data = self._forceplate_data[nplate]['CoP']
        else:
            raise ValueError('Invalid kind of forceplate data requested')
        return self._normalized_analog_data(data)

    """WIP"""
    def get_accelerometer_data(self):
        return read_data.get_accelerometer_data(self.source)

    def _get_fp_events(self):
        """Read the forceplate events."""
        try:
            fp_info = (eclipse.eclipse_fp_keys(self.eclipse_data) if
                       cfg.trial.use_eclipse_fp_info else None)
            # FIXME: marker data already read?
            return utils.detect_forceplate_events(self.source, fp_info=fp_info)
        except GaitDataError:
            logger.warning('Could not detect forceplate events')
            return utils.empty_fp_events()

    def set_norm_cycle(self, cycle=None):
        """ Set normalization cycle.

        The get_ methods will return data normalized to the given cycle.

        Parameters
        ----------
        cycle : int | Gaitcycle | None
            The cycle to normalize data to. Can be a Gaitcycle index (obtain
            cycle by e.g. get_cycles) or a direct index to trial.cycles.
            If None, do not normalize data to gait cycles.
        """

        if isinstance(cycle, int):
            if cycle >= len(self.cycles) or cycle < 0:
                raise ValueError('No such cycle')
            cycle = self.cycles[cycle]
            self._normalize = cycle
        elif isinstance(cycle, Gaitcycle):
            self._normalize = cycle
        elif cycle is None or isinstance(cycle, Noncycle):
            self._normalize = None

    def get_cycles(self, cyclespec):
        """ Get specified gait cycles from the trial as Gaitcycle instances.

        Parameters
        ----------
        cyclespec : dict | str | int | tuple | list
            The cycles to get. Can be dict with 'R' and 'L' keys and
            specification as values to get context specific cycles. If not a
            dict, the given specification will be applied to both contexts.

            'all' gets all trial cycles. 'forceplate' gets cycles starting with
            valid forceplate contact. '1st_forceplate' gets the 1st cycle with
            valid forceplate contact. 'unnormalized' gets a Noncycle that is
            used as a sentinel for unnormalized data.
            An int or a list of int gives the specified cycle indices from the
            trial.
            A tuple can be used to match conditions one by one. For example,
            ('forceplate', 0) would return forceplate cycles if any, and the
            first cycle in case there are none.

        Returns list of gaitcycle instances, sorted by starting frame.
        """
        def _filter_cycles(cycles, context, cyclespec):
            """Takes a list of cycles and filters it according to cyclespec,
            returning only cycles that match the spec"""
            if cyclespec is None:
                return [Noncycle(context=context, trial=self)]
            elif isinstance(cyclespec, int):
                return [cycles[cyclespec]] if cyclespec < len(cycles) else []
            elif isinstance(cyclespec, list):
                return [cycles[c] for c in cyclespec if c < len(cycles)]
            elif cyclespec == 'unnormalized':
                return [Noncycle(context=context)]
            elif cyclespec == 'all':
                return cycles
            elif cyclespec == 'forceplate':  # all forceplate cycles
                return [c for c in cycles if c.on_forceplate]
            elif cyclespec == '1st_forceplate':  # 1st forceplate cycle
                return [c for c in cycles if c.on_forceplate][:1]
            elif isinstance(cyclespec, tuple):
                # recurse until we have cycles (or cyclespec is exhausted)
                if not cyclespec:
                    return []
                else:
                    return (_filter_cycles(cycles, context, cyclespec[0]) or
                            _filter_cycles(cycles, context, cyclespec[1:]))
            else:
                raise ValueError('Invalid argument')

        if not isinstance(cyclespec, dict):
            cyclespec = {'R': cyclespec, 'L': cyclespec}

        cycs_ok = list()
        for context in cyclespec:
            # pick trial cycles for this context
            cycles_ = [c for c in self.cycles if c.context == context.upper()]
            # filter them according to cyclespec
            good_cycles = _filter_cycles(cycles_, context, cyclespec[context])
            cycs_ok.extend(good_cycles)

        return sorted(cycs_ok, key=lambda cyc: cyc.start)

    def _get_modelvar(self, var):
        """Return unnormalized data for a model variable."""
        model_ = models.model_from_var(var)
        if not model_:
            raise ValueError('No model found for %s' % var)
        if model_.desc not in self._models_data:
            # read and cache model data
            modeldata = read_data.get_model_data(self.source, model_)
            self._models_data[model_.desc] = modeldata
        return self._models_data[model_.desc][var]

    def _scan_cycles(self):
        """Create Gaitcycle instances based on trial strike/toeoff markers."""
        # The events marked in the trial marked events need to be matched
        # with detected forceplate events, but may not match exactly, so use
        # a tolerance
        STRIKE_TOL = 7
        sidestrs = {'R': 'right', 'L': 'left'}
        for strikes in [self.lstrikes, self.rstrikes]:
            len_s = len(strikes)
            if len_s < 2:
                continue
            if strikes == self.lstrikes:
                toeoffs = self.ltoeoffs
                context = 'L'
            else:
                toeoffs = self.rtoeoffs
                context = 'R'
            for k in range(0, len_s-1):
                start = strikes[k]
                # see if cycle starts on forceplate strike
                fp_strikes = np.array(self.fp_events[context + '_strikes'])
                if fp_strikes.size == 0:
                    on_forceplate = False
                    plate_idx = None
                else:
                    diffs = np.abs(fp_strikes - start)
                    on_forceplate = min(diffs) <= STRIKE_TOL
                    if on_forceplate:
                        strike_idx = np.argmin(diffs)
                        plate_idx = self.fp_events[context + '_strikes_plate'][strike_idx]
                    else:
                        plate_idx = None
                    logger.debug('side %s: cycle start: %d, '
                                 'detected fp events: %s'
                                 % (context, start, fp_strikes))
                end = strikes[k+1]
                toeoff = [x for x in toeoffs if x > start and x < end]
                if len(toeoff) == 0:
                    raise GaitDataError('%s: no toeoff for cycle starting at '
                                        '%d' % (self.trialname, start))
                elif len(toeoff) > 1:
                    raise GaitDataError('%s: multiple toeoffs for cycle '
                                        'starting at %d' % (self.trialname,
                                                            start))
                else:
                    toeoff = toeoff[0]
                fp_str = ' (fp)' if on_forceplate else ''
                name = '%s%d%s' % (sidestrs[context], (k+1), fp_str)
                yield Gaitcycle(start, end, toeoff, context,
                                on_forceplate, plate_idx, self.samplesperframe,
                                trial=self, index=k+1, name=name)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""

Kin* consistency plots. Automatically picks trials based on Eclipse
description and defined search strings.

@author: Jussi (jnu@iki.fi)
"""

import logging
import argparse
import os.path as op
from itertools import cycle

from gaitutils import (Plotter, cfg, register_gui_exception_handler,
                       GaitDataError, nexus, sessionutils, Trial, plot_plotly)

logger = logging.getLogger(__name__)


def do_plot(sessions=None, tags=None, show=True, make_pdf=True,
            session_styles=False, backend=None):
    """ Find trials according to tags in each session and superpose all.
    By default, plots tagged trials in the current Nexus session.
    session_styles uses different linestyles for each session
    """
    if backend is None:
        backend = cfg.plot.backend

    if sessions is None:
        sessions = [nexus.get_sessionpath()]
    layout = cfg.layouts.overlay_lb_kin

    if tags is None:
        tags = cfg.eclipse.tags

    if backend == 'matplotlib':
        pl = Plotter()
        pl.layout = layout

        linecolors = cfg.plot.overlay_colors
        ccolors = cycle(linecolors)
        linestyles = [':', '--', '-']

        ind = 0
        for session in sessions:
            c3ds = sessionutils.get_c3ds(session, tags=tags,
                                         trial_type='dynamic')
            if not c3ds:
                raise GaitDataError('No marked trials found for session %s'
                                    % session)
            session_style = linestyles.pop()
            for c3d in c3ds:
                pl.open_trial(c3d)
                ind += 1
                if ind > len(linecolors):
                    logger.warning('not enough colors for plot!')
                # only plot normaldata for last trial to speed up things
                plot_model_normaldata = (c3d == c3ds[-1] and
                                         session == sessions[-1])
                # select style/color according to either session or trial
                model_tracecolor = next(ccolors)
                if session_styles:
                    model_linestyle = session_style
                    linestyles_context = False
                else:
                    model_linestyle = None
                    linestyles_context = True

                pl.plot_trial(model_tracecolor=model_tracecolor,
                              model_linestyle=model_linestyle,
                              linestyles_context=linestyles_context,
                              toeoff_markers=False, legend_maxlen=10,
                              maintitle='', superpose=True, show=False,
                              plot_model_normaldata=plot_model_normaldata)

        # auto set title
        if len(sessions) > 1:
            maintitle = 'Kinematics comparison '
            maintitle += ' vs. '.join([op.split(s)[-1] for s in sessions])
        else:
            maintitle = ('Kinematics consistency plot, session %s' %
                         op.split(sessions[0])[-1])
        pl.set_title(maintitle)

        if show:
            pl.show()

        # to recreate old behavior...
        if make_pdf and len(sessions) == 1:
            pl.create_pdf(pdf_name=op.join(sessions[0], 'kin_consistency.pdf'))

        return pl.fig

    elif backend == 'plotly':
        c3ds_all = list()
        for session in sessions:
            c3ds = sessionutils.get_c3ds(session, tags=tags,
                                         trial_type='dynamic')
            if not c3ds:
                raise GaitDataError('No marked trials found for session %s'
                                    % session)
            c3ds_all.extend(c3ds)
        trials = [Trial(c3d) for c3d in c3ds]
        maintitle = ('Kinematics consistency plot, session %s' %
                     op.split(sessions[0])[-1])
        plot_plotly.plot_trials_browser(trials, layout,
                                        legend_type='short_name_with_tag',
                                        maintitle=None)

    else:
        raise ValueError('Invalid backend')


if __name__ == '__main__':

    parser = argparse.ArgumentParser()
    parser.add_argument('--tags', metavar='p', type=str, nargs='+',
                        help='strings that must appear in trial '
                        'description or notes')
    args = parser.parse_args()
    logging.basicConfig(level=logging.DEBUG)
    register_gui_exception_handler()
    do_plot(sessions=[nexus.get_sessionpath()], tags=args.tags)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Plot trial velocities.

@author: Jussi (jnu@iki.fi)
"""

import argparse
from builtins import range
import numpy as np
import plotly.graph_objs as go
import plotly
from scipy import signal
import matplotlib.pyplot as plt
from matplotlib.backends.backend_pdf import PdfPages
import os.path as op
import logging

from gaitutils import (nexus, utils, sessionutils,
                       register_gui_exception_handler)

logger = logging.getLogger(__name__)


def _plot_vel_curves(sessionpath=None):
    """Plot time-dependent velocity for each dynamic trial in session."""
    if sessionpath is None:
        sessionpath = nexus.get_sessionpath()
    c3ds = sessionutils.get_c3ds(sessionpath, trial_type='dynamic')
    if len(c3ds) == 0:
        raise Exception('Did not find any dynamic trials in current '
                        'session directory')
    traces = list()
    for c3d in c3ds[:10]:
        v, vel = utils._trial_median_velocity(c3d, return_curve=True)
        #vel = signal.medfilt(vel, 3)  # if spikes
        tname = op.split(c3d)[-1]
        trace = go.Scatter(y=vel, text=tname, name=tname, hoverinfo='x+y+text')
        traces.append(trace)
    plotly.offline.plot(traces)


def do_plot(sessionpath=None, show=True, make_pdf=True):
    """Plot median velocities for each dynamic trial in Nexus session."""
    if sessionpath is None:
        sessionpath = nexus.get_sessionpath()
    c3ds = sessionutils.get_c3ds(sessionpath, trial_type='dynamic')

    if len(c3ds) == 0:
        raise Exception('Did not find any dynamic trials in current '
                        'session directory')

    labels = [op.splitext(op.split(f)[1])[0] for f in c3ds]
    vels = np.array([utils._trial_median_velocity(trial) for trial in c3ds])
    vavg = np.nanmean(vels)

    fig = plt.figure()
    plt.stem(vels)
    plt.xticks(range(len(vels)), labels, rotation='vertical')
    plt.ylabel('Speed (m/s)')
    plt.tick_params(axis='both', which='major', labelsize=8)
    plt.title('Walking speed for dynamic trials (average %.2f m/s)' % vavg)
    plt.tight_layout()

    if make_pdf:
        pdf_name = op.join(nexus.get_sessionpath(), 'trial_velocity.pdf')
        with PdfPages(pdf_name) as pdf:
            pdf.savefig(fig)

    if show:
        plt.show()

    return fig


if __name__ == '__main__':
    logging.basicConfig(level=logging.DEBUG)
    register_gui_exception_handler()
    parser = argparse.ArgumentParser()
    parser.add_argument('--curves', action='store_true',
                        help='plot velocity curves')
    args = parser.parse_args()
    if args.curves:
        _plot_vel_curves()
    else:
        do_plot()



# -*- coding: utf-8 -*-
"""

Utility functions for processing gait data.

@author: Jussi (jnu@iki.fi)
"""

from __future__ import division

from builtins import str
from builtins import zip
from scipy import signal
from matplotlib import path
import matplotlib.pyplot as plt
import numpy as np
import logging

from .envutils import GaitDataError
from .numutils import (rising_zerocross, best_match, falling_zerocross,
                       _baseline)
from .config import cfg

logger = logging.getLogger(__name__)


def get_crossing_frame(mP, dim=1, p0=0):
    """ Return frame(s) where marker position (dimension dim) crosses p0
    (units are as returned by Nexus, usually mm).
    Dims are x=0, y=1, z=2. """
    y = mP[:, dim]
    nzind = np.where(y != 0)  # nonzero elements == valid data (not nice)
    y[nzind] -= p0
    zx = np.append(rising_zerocross(y), falling_zerocross(y))
    ycross = list()
    # sanity checks
    for p in zx:
        # y must be nonzero on either side of crossing (valid data)
        if p-10 > 0 and p+10 < len(y):
            if y[p-10] != 0 and y[p+10] != 0:
                # y must change sign also around p
                if np.sign(y[p-10]) != np.sign(y[p+10]):
                        ycross.append(p)
    return ycross


def avg_markerdata(mkrdata, markers, var_type='_P', roi=None):
    """ Average marker data.
    """
    data_shape = mkrdata[markers[0]+var_type].shape
    mP = np.zeros(data_shape)
    if roi is None:
        roi = [0, data_shape[0]]
    roi_frames = np.arange(roi[0], roi[1])
    n_ok = 0
    for marker in markers:
        gap_frames = mkrdata[marker+'_gaps']
        if np.intersect1d(roi_frames, gap_frames).size > 0:
            raise GaitDataError('Averaging data for %s has gaps' % marker)
        else:
            mP += mkrdata[marker+var_type]
            n_ok += 1
    else:
        return mP / n_ok


# FIXME: marker sets could be moved into models.py?
def _pig_markerset(fullbody=True, sacr=True):
    """ PiG marker set as dict (empty values) """
    _pig = ['LASI', 'RASI', 'LTHI', 'LKNE', 'LTIB', 'LANK', 'LHEE',
            'LTOE', 'RTHI', 'RKNE', 'RTIB', 'RANK', 'RHEE', 'RTOE']
    if fullbody:
        _pig += ['LFHD', 'RFHD', 'LBHD', 'RBHD', 'C7', 'T10', 'CLAV', 'STRN',
                 'RBAK', 'LSHO', 'LELB', 'LWRA', 'LWRB', 'LFIN', 'RSHO',
                 'RELB', 'RWRA', 'RWRB', 'RFIN']
    # add pelvis posterior markers; SACR or RPSI/LPSI
    _pig.extend(['SACR'] if sacr else ['RPSI', 'LPSI'])
    return {mkr: None for mkr in _pig}


def _pig_pelvis_markers():
    return ['RASI', 'RPSI', 'LASI', 'LPSI', 'SACR']


def is_plugingait_set(mkrdata):
    """ Check whether marker data set corresponds to Plug-in Gait (full body or
    lower body only). Extra markers are accepted. """
    mkrs = set(mkrdata.keys())
    # required markers
    lb_mkrs_sacr = set(_pig_markerset(fullbody=False).keys())
    lb_mkrs_psi = set(_pig_markerset(fullbody=False, sacr=False).keys())
    return lb_mkrs_psi.issubset(mkrs) or lb_mkrs_sacr.issubset(mkrs)


def _check_markers_flipped(mkrdata):
    """Checks for markers that may typically get flipped in labeling.
    Yields pairs of flipped markers"""
    MAX_ANGLE = 90  # max angle to consider vectors 'similarly oriented'
    # HEE-TOE checks
    for side in ['L', 'R']:
        # compare HEE-TOE line to pelvis orientation
        mkr_toe, mkr_hee = '%sTOE' % side, '%sHEE' % side
        ht = _normalize(mkrdata[mkr_toe] - mkrdata[mkr_hee])
        if side+'PSI' in mkrdata:
            pa = _normalize(mkrdata[side+'ASI'] - mkrdata[side+'PSI'])
        elif 'SACR' in mkrdata:
            pa = _normalize(mkrdata[side+'ASI'] - mkrdata['SACR'])
        angs = np.arccos(np.sum(ht * pa, axis=1)) / np.pi * 180
        if np.nanmedian(angs) > MAX_ANGLE:
            yield mkr_toe, mkr_hee


def principal_movement_direction(mP):
    """ Return principal movement direction (dimension of maximum variance) """
    inds_ok = np.where(np.any(mP, axis=1))  # make sure that gaps are ignored
    return np.argmax(np.var(mP[inds_ok], axis=0))


def get_foot_contact_velocity(mkrdata, fp_events, medians=True, roi=None):
    """ Return foot velocities during forceplate strike/toeoff frames.
    fp_events is from detect_forceplate_events()
    If medians=True, return median values. """
    results = dict()
    for context, markers in zip(('R', 'L'), [cfg.autoproc.right_foot_markers,
                                cfg.autoproc.left_foot_markers]):
        footctrv_ = avg_markerdata(mkrdata, markers, var_type='_V', roi=roi)
        footctrv = np.linalg.norm(footctrv_, axis=1)
        strikes = fp_events[context+'_strikes']
        toeoffs = fp_events[context+'_toeoffs']
        results[context + '_strike'] = footctrv[strikes]
        results[context + '_toeoff'] = footctrv[toeoffs]
    if medians:
        results = {key: (np.array([np.median(x)]) if x.size > 0 else x)
                   for key, x in results.items()}
    return results


def _get_foot_swing_velocity(footctrv, max_peak_velocity, min_swing_velocity):
    """Compute foot swing velocity from scalar velocity data (footctrv)"""
    # find maxima of velocity: derivative crosses zero and values ok
    vd = np.gradient(footctrv)
    vdz_ind = falling_zerocross(vd)
    inds = np.where(footctrv[vdz_ind] < max_peak_velocity)[0]
    if len(inds) == 0:
        raise GaitDataError('Cannot find acceptable velocity peaks')
    # delete spurious peaks (where min swing velocity is not attained)
    vs = footctrv[vdz_ind[inds]]
    not_ok = np.where(vs < vs.max() * min_swing_velocity)
    vs = np.delete(vs, not_ok)
    return np.median(vs)


def _normalize(V):
    """Normalize rows of matrix V"""
    Vn = np.linalg.norm(V, axis=1)
    # quietly return all nans for length zero vectors
    with np.errstate(divide='ignore', invalid='ignore'):
        return V / Vn[:, np.newaxis]


def _get_foot_points(mkrdata, context, footlen=None):
    """Estimate points in the xy plane enclosing the foot. Foot is modeled
    as a triangle"""
    # marker data as N x 3 matrices
    heeP = mkrdata[context+'HEE']
    toeP = mkrdata[context+'TOE']
    ankP = mkrdata[context+'ANK']
    # heel - toe vectors
    htV = toeP - heeP
    htVn = _normalize(htV)
    # heel - ankle vectors
    haV = ankP - heeP
    ha_len = np.linalg.norm(haV, axis=1)
    # end point of foot (just beyond 2nd toe)
    if footlen is None:
        # rough estimate based on marker distances
        logger.debug('using estimated foot length')
        footlen = np.median(ha_len) * cfg.autoproc.foot_relative_len
    foot_end = heeP + htVn * footlen
    # projection of HEE-ANK to HEE-TOE line
    ha_htV = htVn * np.sum(haV*htVn, axis=1)[:, np.newaxis]
    # lateral ANK vector (HEE-TOE line to ANK)
    lankV = haV - ha_htV
    # edge points are coplanar with markers but not with the foot
    # lateral foot edge
    lat_edge = foot_end + lankV
    # medial foot edge
    med_edge = foot_end - lankV
    # heel edge (compensate for marked position)
    heel_edge = heeP + htVn * cfg.autoproc.marker_diam/2
    logger.debug('foot length: %.1f mm width: %.1f mm' %
                 (np.nanmedian(np.linalg.norm(heel_edge-foot_end, axis=1)),
                  np.nanmedian(np.linalg.norm(lat_edge-med_edge, axis=1))))
    return {'heel': heel_edge, 'lateral': lat_edge, 'medial': med_edge,
            'toe': foot_end}


def _leading_foot(mkrdata, roi=None):
    """Determine which foot is leading (ahead in the direction of gait).
    Returns n-length list of 'R' or 'L' correspondingly (n = number of
    frames). Gaps are indicated as None. mkrdata must include foot and
    pelvis markers"""
    subj_pos = avg_markerdata(mkrdata, cfg.autoproc.track_markers, roi=roi)
    # FIXME: should not use a single dim here
    gait_dim = principal_movement_direction(subj_pos)
    gait_dir = np.median(np.diff(subj_pos, axis=0), axis=0)[gait_dim]
    lfoot = avg_markerdata(mkrdata, cfg.autoproc.left_foot_markers,
                           roi=roi)[:, gait_dim]
    rfoot = avg_markerdata(mkrdata, cfg.autoproc.right_foot_markers,
                           roi=roi)[:, gait_dim]
    cmpfun = np.greater if gait_dir > 0 else np.less
    return [None if R == 0.0 or L == 0.0 else ('R' if cmpfun(R, L) else 'L')
            for R, L in zip(rfoot, lfoot)]


def _trial_median_velocity(source, return_curve=False):
    """Compute median velocity (walking speed) over whole trial by
    differentiation of marker data from track markers. Up/down movement of
    markers may slightly increase speed compared to time-distance values.
    If return_curve, return velocity curve normalized to 0..100% of trial"""
    from . import read_data
    try:
        frate = read_data.get_metadata(source)['framerate']
        mkrdata = read_data.get_marker_data(source, cfg.autoproc.track_markers)
        vel_3 = avg_markerdata(mkrdata, cfg.autoproc.track_markers,
                               var_type='_V')
        vel_ = np.sqrt(np.sum(vel_3**2, 1))  # scalar velocity
    except (GaitDataError, ValueError):
        if return_curve:
            nanvec = np.empty((100, 1))
            nanvec[:] = np.nan
            return np.nan, nanvec
        else:
            return np.nan
    vel = np.median(vel_[np.where(vel_)])  # ignore zeros
    vel_ms = vel * frate / 1000.  # convert to m/s
    if return_curve:
        tn = np.linspace(0, 100, 101)
        vel_curve = np.interp(tn, np.linspace(0, 100, len(vel_)), vel_)
        return vel_ms, vel_curve * frate / 1000.
    else:
        return vel_ms


def _point_in_poly(poly, pt):
    """Point-in-polygon. poly is ordered nx3 array of vertices and P is mx3
    array of points. Returns mx3 array of booleans. 3rd dim is currently
    ignored"""
    p = path.Path(poly[:, :2])
    return p.contains_point(pt)


def empty_fp_events():
    """Helper to return no forceplate events"""
    return dict(R_strikes=[], R_toeoffs=[], L_strikes=[], L_toeoffs=[],
                valid=set(), R_strikes_plate=[], L_strikes_plate=[],
                our_fp_info={}, coded='')


def detect_forceplate_events(source, mkrdata=None, fp_info=None,
                             roi=None):
    """ Detect frames where valid forceplate strikes and toeoffs occur.
    Uses forceplate data and estimated foot shape.

    If supplied, mkrdata must include foot and pelvis markers. Otherwise
    it will be read.

    If fp_info dict is supplied, no marker-based checks will be done;
    instead the Eclipse forceplate info will be used to determine the foot.
    Eclipse info is written e.g. as {FP1: 'Left'} where plate indices start
    from 1 and the value can be 'Auto', 'Left', 'Right' or 'Invalid'.
    Even if Eclipse info is used, foot strike and toeoff frames must be
    determined from forceplate data.

    If roi is given e.g. [100, 300], all marker data checks will be restricted
    to roi.
    """
    def _foot_plate_check(fpdata, mkrdata, fr0, side, footlen):
        """Helper for foot-plate check. Returns 0, 1, 2 for:
            completely outside plate, partially outside plate, inside plate"""
        allpts = _get_foot_points(mkrdata, side, footlen)
        poly = fpdata['cor_full']
        pts_ok = list()
        for label, pts in allpts.items():
            pt = pts[fr0, :]
            pt_ok = _point_in_poly(poly, pt)
            logger.debug('%s point %son plate' %
                         (label, '' if pt_ok else 'not '))
            pts_ok.append(pt_ok)
        if all(pts_ok):
            return 2
        elif any(pts_ok):
            return 1
        else:
            return 0

    # get subject info
    from . import read_data
    logger.debug('detect forceplate events from %s' % source)
    info = read_data.get_metadata(source)
    fpdata = read_data.get_forceplate_data(source)
    results = empty_fp_events()

    # get marker data and find "forward" direction (by max variance)
    if mkrdata is None:
        mkrs = (cfg.autoproc.right_foot_markers +
                cfg.autoproc.left_foot_markers +
                cfg.autoproc.track_markers)
        mkrdata = read_data.get_marker_data(source, mkrs)

    footlen = info['subj_params']['FootLen']
    rfootlen = info['subj_params']['RFootLen']
    lfootlen = info['subj_params']['LFootLen']
    if footlen is not None:
        logger.debug('(obsolete) single foot length parameter set to %.2f'
                     % footlen)
        rfootlen = lfootlen = footlen
    elif rfootlen is not None and lfootlen is not None:
        logger.debug('foot length parameters set to r=%.2f, l=%.2f' %
                     (rfootlen, lfootlen))
    else:
        logger.debug('foot length parameter not set')
    bodymass = info['subj_params']['Bodymass']

    logger.debug('acquiring gait events')
    events_0 = automark_events(source, mkrdata=mkrdata, mark=False,
                               roi=roi)

    # loop over plates; our internal forceplate index is 0-based
    for plate_ind, fp in enumerate(fpdata):
        logger.debug('analyzing plate %d' % plate_ind)
        # XXX: are we sure that the plate indices always match Eclipse?
        plate = 'FP' + str(plate_ind + 1)  # Eclipse starts from FP1
        if fp_info is not None and plate in fp_info:
            ecl_valid = fp_info[plate]
            detect_foot = False
            logger.debug('using Eclipse forceplate info: %s' % ecl_valid)
            if ecl_valid == 'Right':
                valid = 'R'
            elif ecl_valid == 'Left':
                valid = 'L'
            elif ecl_valid == 'Invalid':
                valid = None
            elif ecl_valid == 'Auto':
                detect_foot = True
            else:
                raise GaitDataError('unexpected Eclipse forceplate field')
        else:
            logger.debug('not using Eclipse forceplate info')
            valid = None
            detect_foot = True

        # identify candidate frames for foot strike
        # FIXME: filter should maybe depend on sampling freq
        force_checks_ok = True
        forcetot = signal.medfilt(fp['Ftot'])
        forcetot = _baseline(forcetot)
        fmax = max(forcetot)
        fmaxind = np.where(forcetot == fmax)[0][0]  # first maximum
        logger.debug('max force: %.2f N at %.2f' % (fmax, fmaxind))

        if bodymass is None:
            f_threshold = cfg.autoproc.forceplate_contact_threshold * fmax
            logger.warning('body mass unknown, thresholding force at %.2f N',
                           f_threshold)
        else:
            logger.debug('body mass %.2f kg' % bodymass)
            f_threshold = (cfg.autoproc.forceplate_contact_threshold *
                           bodymass * 9.81)
            if fmax < cfg.autoproc.forceplate_min_weight * bodymass * 9.81:
                logger.debug('insufficient max. force on plate')
                force_checks_ok = False

        # find indices where force crosses threshold
        try:
            logger.debug('force threshold: %.2f N' % f_threshold)
            friseind = rising_zerocross(forcetot-f_threshold)[0]  # first rise
            ffallind = falling_zerocross(forcetot-f_threshold)[-1]  # last fall
            logger.debug('force rise: %d fall: %d' % (friseind, ffallind))
            # 0-based frame indices (=1 less than Nexus frame index)
            strike_fr = int(np.round(friseind / info['samplesperframe']))
            toeoff_fr = int(np.round(ffallind / info['samplesperframe']))
            logger.debug('strike @ frame %d, toeoff @ %d'
                         % (strike_fr, toeoff_fr))
        except IndexError:
            logger.debug('cannot detect force rise/fall')
            force_checks_ok = False

        if not force_checks_ok:
            valid = None

        # foot marker (or point) checks
        if force_checks_ok and detect_foot:
            logger.debug('using autodetection of foot contact')
            # allows foot to settle for 50 ms after strike
            settle_fr = int(50/1000 * info['framerate'])
            fr0 = strike_fr + settle_fr
            side = _leading_foot(mkrdata, roi=roi)[fr0]
            if side is None:
                raise GaitDataError('cannot determine leading foot from marker'
                                    ' data')
            footlen = rfootlen if side == 'R' else lfootlen
            logger.debug('checking contact for leading foot: %s' % side)
            ok = _foot_plate_check(fp, mkrdata, fr0, side, footlen) == 2
            # check that contralateral foot is not on plate (needs events)
            if ok and events_0 is not None:
                contra_side = 'R' if side == 'L' else 'L'
                contra_strikes = events_0[contra_side+'_strikes']
                contra_strikes_next = contra_strikes[np.where(contra_strikes >
                                                              strike_fr)]
                if contra_strikes_next.size == 0:
                    logger.debug('no following contralateral strike')
                else:
                    fr0 = contra_strikes_next[0] + settle_fr
                    logger.debug('checking next contact for contralateral '
                                 'foot (at frame %d)' % fr0)
                    contra_next_ok = _foot_plate_check(fp, mkrdata, fr0,
                                                       contra_side,
                                                       footlen) == 0
                    ok &= contra_next_ok
                contra_strikes_prev = contra_strikes[np.where(contra_strikes <
                                                              strike_fr)]
                if contra_strikes_prev.size == 0:
                    logger.debug('no previous contralateral strike')
                else:
                    fr0 = contra_strikes_prev[-1] + settle_fr
                    logger.debug('checking previous contact for contralateral '
                                 'foot (at frame %d)' % fr0)
                    contra_prev_ok = _foot_plate_check(fp, mkrdata, fr0,
                                                       contra_side,
                                                       footlen) == 0
                    ok &= contra_prev_ok
            valid = side if ok else None

        if valid:
            logger.debug('plate %d: valid foot strike on %s at frame %d'
                         % (plate_ind, valid, strike_fr))
            results['valid'].add(valid)
            results[valid+'_strikes'].append(strike_fr)
            results[valid+'_toeoffs'].append(toeoff_fr)
            results[valid+'_strikes_plate'].append(plate_ind)
            results['our_fp_info'][plate] = 'Right' if valid == 'R' else 'Left'
            results['coded'] += valid
        else:
            logger.debug('plate %d: no valid foot strike' % plate_ind)
            results['our_fp_info'][plate] = 'Invalid'
            results['coded'] += 'X'

    logger.debug(results)
    return results


def automark_events(source, mkrdata=None, events_range=None, fp_events=None,
                    vel_thresholds=None, roi=None,
                    start_on_forceplate=False, plot=False, mark=True):

    """ Mark events based on velocity thresholding. Absolute thresholds
    can be specified as arguments. Otherwise, relative thresholds will be
    calculated based on the data. Optimal results will be obtained when
    thresholds based on force plate data are precomputed.

    If mkrdata is None, it will be read from source. Otherwise mkrdata must
    include both foot markers and the body tracking markers (see config)

    vel_thresholds gives velocity thresholds for identifying events. These
    can be obtained from forceplate data (utils.check_forceplate_contact).
    Separate thresholds for left and right side.

    fp_events is dict specifying the forceplate detected strikes and toeoffs
    (see utils.detect_forceplate_events). These will be used instead of
    (nearby) autodetected events.

    If events_range is specified, the events will be restricted to given
    coordinate range in the principal gait direction.
    E.g. events_range=[-1000, 1000]
    
    If roi is specified, events will be restricted to roi.

    If start_on_forceplate is True, the first cycle will start on forceplate
    (i.e. events earlier than the first foot strike events in fp_events will
    not be marked for the corresponding side(s)).

    If plot=True, velocity curves and events are plotted.

    If mark=False, no events will actually be marked in Nexus.

    Before automark, run reconstruct, label, gap fill and filter pipelines.
    Filtering is important to get reasonably smooth derivatives.
    """

    from .read_data import get_metadata, get_marker_data
    info = get_metadata(source)
    frate = info['framerate']
    # some operations are Nexus specific and make no sense for c3d source
    from . import nexus

    # TODO: move into config
    # marker data is assumed to be in mm
    # mm/frame = 1000 m/frame = 1000/frate m/s
    VEL_CONV = 1000/frate
    # reasonable limit for peak velocity (m/s before multiplier)
    MAX_PEAK_VELOCITY = 12 * VEL_CONV
    # reasonable limits for velocity on slope (increasing/decreasing)
    MAX_SLOPE_VELOCITY = 6 * VEL_CONV
    MIN_SLOPE_VELOCITY = 0  # not currently in use
    # minimum swing velocity (rel to max velocity)
    MIN_SWING_VELOCITY = .5
    # median prefilter width
    PREFILTER_MEDIAN_WIDTH = 3
    # tolerance for matching forceplate and vel. thresholded events
    FP_EVENT_TOL = 10

    if vel_thresholds is None:
        vel_thresholds = {'L_strike': None, 'L_toeoff': None,
                          'R_strike': None, 'R_toeoff': None}

    if mkrdata is None:
        reqd_markers = (cfg.autoproc.right_foot_markers +
                        cfg.autoproc.left_foot_markers +
                        cfg.autoproc.track_markers)
        mkrdata = get_marker_data(source, reqd_markers)

    rfootctrv_ = avg_markerdata(mkrdata, cfg.autoproc.right_foot_markers,
                                var_type='_V', roi=roi)
    rfootctrv = np.linalg.norm(rfootctrv_, axis=1)
    lfootctrv_ = avg_markerdata(mkrdata, cfg.autoproc.left_foot_markers,
                                var_type='_V', roi=roi)
    lfootctrv = np.linalg.norm(lfootctrv_, axis=1)

    # position data: use ANK marker
    rfootctrP = mkrdata['RANK']
    lfootctrP = mkrdata['LANK']

    strikes_all = {}
    toeoffs_all = {}

    # loop: same operations for left / right foot
    for context, footctrP, footctrv in zip(('R', 'L'), (rfootctrP, lfootctrP),
                                           (rfootctrv, lfootctrv)):
        logger.debug('marking side %s' % context)
        # foot center position
        # filter scalar velocity data to suppress noise and spikes
        footctrv = signal.medfilt(footctrv, PREFILTER_MEDIAN_WIDTH)
        # get peak (swing) velocity
        maxv = _get_foot_swing_velocity(footctrv, MAX_PEAK_VELOCITY,
                                        MIN_SWING_VELOCITY)

        # compute thresholds
        if (cfg.autoproc.use_fp_vel_thresholds and
           vel_thresholds[context+'_strike']):
            threshold_fall_ = vel_thresholds[context+'_strike']
        else:
            threshold_fall_ = maxv * cfg.autoproc.strike_vel_threshold
        if (cfg.autoproc.use_fp_vel_thresholds and
           vel_thresholds[context+'_toeoff']):
            threshold_rise_ = vel_thresholds[context+'_toeoff']
        else:
            threshold_rise_ = maxv * cfg.autoproc.toeoff_vel_threshold
        logger.debug('side: %s, default thresholds fall/rise: %.2f/%.2f'
                     % (context, maxv * cfg.autoproc.strike_vel_threshold,
                        maxv * cfg.autoproc.toeoff_vel_threshold))
        logger.debug('using thresholds: %.2f/%.2f' % (threshold_fall_,
                                                      threshold_rise_))
        # find point where velocity crosses threshold
        # foot strikes (velocity decreases)
        cross = falling_zerocross(footctrv - threshold_fall_)
        # exclude edges of data vector
        fmax = len(footctrv) - 1
        cross = cross[np.where(np.logical_and(cross > 0, cross < fmax))]
        # check velocity on slope
        cind_min = np.logical_and(footctrv[cross-1] < MAX_SLOPE_VELOCITY,
                                  footctrv[cross-1] > MIN_SLOPE_VELOCITY)
        cind_max = np.logical_and(footctrv[cross+1] < MAX_SLOPE_VELOCITY,
                                  footctrv[cross+1] > MIN_SLOPE_VELOCITY)
        strikes = cross[np.logical_and(cind_min, cind_max)]

        # check for foot swing (velocity maximum) between consecutive strikes
        # if no swing, keep deleting the latter event until swing is found
        bad = []
        for sind in range(len(strikes)):
            if sind in bad:
                continue
            for sind2 in range(sind+1, len(strikes)):
                swing_max_vel = footctrv[strikes[sind]:strikes[sind2]].max()
                # logger.debug('check %d-%d' % (strikes[sind], strikes[sind2]))
                if swing_max_vel < maxv * MIN_SWING_VELOCITY:
                    logger.debug('no swing between strikes %d-%d, deleting %d'
                                 % (strikes[sind], strikes[sind2],
                                    strikes[sind2]))
                    bad.append(sind2)
                else:
                    break
        strikes = np.delete(strikes, bad)

        if len(strikes) == 0:
            raise GaitDataError('No valid foot strikes detected')

        # toe offs (velocity increases)
        cross = rising_zerocross(footctrv - threshold_rise_)
        cross = cross[np.where(np.logical_and(cross > 0,
                                              cross < len(footctrv)))]
        cind_min = np.logical_and(footctrv[cross-1] < MAX_SLOPE_VELOCITY,
                                  footctrv[cross-1] > MIN_SLOPE_VELOCITY)
        cind_max = np.logical_and(footctrv[cross+1] < MAX_SLOPE_VELOCITY,
                                  footctrv[cross+1] > MIN_SLOPE_VELOCITY)
        toeoffs = cross[np.logical_and(cind_min, cind_max)]

        if len(toeoffs) == 0:
            raise GaitDataError('Could not detect any toe-off events')

        # check for multiple toeoffs
        for s1, s2 in list(zip(strikes, np.roll(strikes, -1)))[:-1]:
            to_this = np.where(np.logical_and(toeoffs > s1, toeoffs < s2))[0]
            if len(to_this) > 1:
                logger.debug('%d toeoffs during cycle, keeping the last one'
                             % len(to_this))
                toeoffs = np.delete(toeoffs, to_this[:-1])

        logger.debug('autodetected strike events: %s' % strikes)
        logger.debug('autodetected toeoff events: %s' % toeoffs)

        # select events for which the foot is close enough to center frame
        if events_range:
            mP = avg_markerdata(mkrdata, cfg.autoproc.track_markers, roi=roi)
            fwd_dim = principal_movement_direction(mP)
            strike_pos = footctrP[strikes, fwd_dim]
            dist_ok = np.logical_and(strike_pos > events_range[0],
                                     strike_pos < events_range[1])
            # exactly zero position at strike should indicate a gap -> exclude
            # TODO: smarter gap handling
            dist_ok = np.logical_and(dist_ok, strike_pos != 0)
            strikes = strikes[dist_ok]

        # correct foot strikes with force plate autodetected events
        if fp_events and fp_events[context+'_strikes']:
            fp_strikes = fp_events[context+'_strikes']
            logger.debug('forceplate strikes: %s' % fp_strikes)
            # find best fp matches for all strikes
            fpc = best_match(strikes, fp_strikes)
            ok_ind = np.where(np.abs(fpc - strikes) < FP_EVENT_TOL)[0]
            if ok_ind.size == 0:
                logger.warning('could not match forceplate strike with an '
                               'autodetected strike')
            else:
                # replace with fp detected strikes
                strikes[ok_ind] = fpc[ok_ind]
                logger.debug('fp corrected strikes: %s' % strikes)
            # toeoffs
            fp_toeoffs = fp_events[context+'_toeoffs']
            logger.debug('forceplate toeoffs: %s' % fp_toeoffs)
            fpc = best_match(toeoffs, fp_toeoffs)
            ok_ind = np.where(np.abs(fpc - toeoffs) < FP_EVENT_TOL)[0]
            if ok_ind.size == 0:
                logger.warning('could not match forceplate toeoff with an '
                               'autodetected toeoff')
            else:
                toeoffs[ok_ind] = fpc[ok_ind]
                logger.debug('fp corrected toeoffs: %s' % toeoffs)
            # delete strikes before (actual) 1st forceplate contact
            if start_on_forceplate and len(fp_strikes) > 0:
                # use a tolerance here to avoid deleting possible (uncorrected)
                # strike near the fp
                not_ok = np.where(strikes < fp_strikes[0] - FP_EVENT_TOL)[0]
                if not_ok.size > 0:
                    logger.debug('deleting foot strikes before forceplate: %s'
                                 % strikes[not_ok])
                    strikes = np.delete(strikes, not_ok)

        if roi is not None:
            strikes = np.extract(np.logical_and(roi[0] <= strikes+1,
                                                strikes+1 <= roi[1]), strikes)

            toeoffs = np.extract(np.logical_and(roi[0] <= toeoffs+1,
                                                toeoffs+1 <= roi[1]), toeoffs)

        if len(strikes) == 0:
            raise GaitDataError('No valid foot strikes detected')

        # delete toeoffs that are not between strike events
        not_ok = np.where(np.logical_or(toeoffs <= min(strikes),
                                        toeoffs >= max(strikes)))
        toeoffs = np.delete(toeoffs, not_ok)

        logger.debug('final strike events: %s' % strikes)
        logger.debug('final toeoff events: %s' % toeoffs)

        if mark:
            if not nexus.is_vicon_instance(source):
                raise ValueError('event marking supported only for Nexus')
            vicon = nexus.viconnexus()
            nexus.create_events(vicon, context, strikes, toeoffs)

        strikes_all[context] = strikes
        toeoffs_all[context] = toeoffs

        # plot velocities w/ thresholds and marked events
        if plot:
            first_call = context == 'R'
            if first_call:
                f, (ax1, ax2) = plt.subplots(2, 1)
            ax = ax1 if first_call else ax2
            ax.plot(footctrv, 'g', label='foot center velocity ' + context)
            # algorithm, fixed thresholds
            ax.plot(strikes, footctrv[strikes], 'kD', markersize=10,
                    label='strike')
            ax.plot(toeoffs, footctrv[toeoffs], 'k^', markersize=10,
                    label='toeoff')
            ax.legend(numpoints=1, fontsize=10)
            ax.set_ylim(0, maxv+10)
            if not first_call:
                plt.xlabel('Frame')
            ax.set_ylabel('Velocity (mm/frame)')
            ax.set_title('Left' if context == 'L' else 'Right')

    if plot:
        plt.show()

    return {'R_strikes': strikes_all['R'], 'L_strikes': strikes_all['L'],
            'R_toeoffs': toeoffs_all['R'], 'L_toeoffs': toeoffs_all['L']}



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Launch gaitutils menu.

@author: Jussi (jnu@iki.fi)
"""
from __future__ import print_function


from gaitutils.gui import _gaitmenu

_gaitmenu.main()



# -*- coding: utf-8 -*-
"""

c3d reader functions


@author: Jussi (jnu@iki.fi)
"""

from __future__ import print_function, division
from builtins import zip
from collections import defaultdict
import logging
import numpy as np
import os

from .numutils import center_of_pressure, change_coords
from .envutils import GaitDataError


logger = logging.getLogger(__name__)
try:
    import btk
except ImportError:
    print('Cannot find btk module; unable to read .c3d files')


def is_c3dfile(obj):
    """ Check whether obj is a valid c3d file. Currently just checks
    existence. """
    try:
        return os.path.isfile(obj)
    except TypeError:
        return False


def _get_c3d_metadata_subfields(acq, field):
    """ Return names of metadata subfields for a given field"""
    meta = acq.GetMetaData()
    meta_s = meta.GetChild(field)
    return [f.GetLabel() for f in btk.Iterate(meta_s)]


def _get_c3d_metadata_field(acq, field, subfield):
    """Get c3d metadata FIELD:SUBFIELD as Python type.
    Always returns a list - pick [0] if scalar """
    meta = acq.GetMetaData()

    def _get_child(field, child):
        try:
            return field.GetChild(child)
        except RuntimeError:
            raise ValueError('Invalid c3d metadata field: %s' % child)

    info = _get_child(_get_child(meta, field), subfield).GetInfo()
    if info.GetFormatAsString() == 'Char':
        return [s.strip() for s in info.ToString()]
    elif info.GetFormatAsString() == 'Real':
        return [x for x in info.ToDouble()]
    else:
        raise ValueError('Unhandled btk meta info type')


def _get_c3dacq(c3dfile):
    """Get a btk.btkAcquisition object"""
    reader = btk.btkAcquisitionFileReader()
    # btk interface cannot take unicode, so encode to latin-1 first
    c3dfile = c3dfile.encode('latin-1')
    reader.SetFilename(c3dfile)
    reader.Update()
    return reader.GetOutput()


def get_analysis(c3dfile, condition='unknown'):
    """Get analysis values from c3d (e.g. gait parameters). Returns a dict
    keyed by var and context. First key can optionally be a condition label."""
    logger.debug('getting analysis values from %s' % c3dfile)
    acq = _get_c3dacq(c3dfile)
    vars_ = _get_c3d_metadata_field(acq, 'ANALYSIS', 'NAMES')
    units = _get_c3d_metadata_field(acq, 'ANALYSIS', 'UNITS')
    contexts = _get_c3d_metadata_field(acq, 'ANALYSIS', 'CONTEXTS')
    vals = _get_c3d_metadata_field(acq, 'ANALYSIS', 'VALUES')

    # build a nice output dict
    di = dict()
    di[condition] = dict()
    di_ = di[condition]

    for (var, unit, context, val) in zip(vars_, units, contexts, vals):
        if var not in di_:
            di_[var] = dict()
            di_[var]['unit'] = unit
        if context not in di_[var]:
            di_[var][context] = val

    for var in di_:
        for context in ['Left', 'Right']:
            if context not in di_[var]:
                logger.warning('%s has missing value: %s / %s' %
                               (c3dfile, var, context))
                di_[var][context] = np.NaN

    return di


def get_emg_data(c3dfile):
    """ Read EMG data from a c3d file. """
    return _get_analog_data(c3dfile, 'EMG')


def get_accelerometer_data(c3dfile):
    """ Read accelerometer data from a c3d file. """
    data = _get_analog_data(c3dfile, 'Accelerometer')
    # Remove the 'Acceleration.' prefix if inserted by Nexus, so that channel
    # names match Nexus. This is a bit ugly (not done for EMG which uses fuzzy
    # matching)
    for key in data['data']:
        if key.find('Acceleration.') == 0:
            # replace key names
            data['data'][key[13:]] = data['data'].pop(key)
    return data


def _get_analog_data(c3dfile, devname):
    """ Read analog data from a c3d file. devname is matched against channel
    names. """
    acq = _get_c3dacq(c3dfile)
    data = dict()
    chnames = []
    for i in btk.Iterate(acq.GetAnalogs()):
        if i.GetDescription().find(devname) >= 0:
            chname = i.GetLabel()
            chnames.append(chname)
            data[chname] = np.squeeze(i.GetValues())
    if chnames:
        return {'t': np.arange(len(data[chname])) / acq.GetAnalogFrequency(),
                'data': data}
    else:
        raise GaitDataError('No matching analog channels found in data')


def _get_marker_data(c3dfile, markers, ignore_edge_gaps=True,
                     ignore_missing=False):
    """Get position, velocity and acceleration for specified markers."""
    if not isinstance(markers, list):  # listify if not already a list
        markers = [markers]
    acq = _get_c3dacq(c3dfile)
    mkrdata = dict()
    for marker in markers:
        try:
            mP = np.squeeze(acq.GetPoint(marker).GetValues())
        except RuntimeError:
            if ignore_missing:
                logger.warning('Cannot read trajectory %s from c3d file'
                               % marker)
                continue
            else:
                raise GaitDataError('Cannot read trajectory %s from c3d file'
                                    % marker)
        mkrdata[marker] = mP
        mkrdata[marker + '_P'] = mP
        mkrdata[marker + '_V'] = np.gradient(mP)[0]
        mkrdata[marker + '_A'] = np.gradient(mkrdata[marker+'_V'])[0]
        # find gaps
        allzero = np.any(mP, axis=1).astype(int)
        if ignore_edge_gaps:
            nleading = allzero.argmax()
            allzero_trim = np.trim_zeros(allzero)
            gap_inds = np.where(allzero_trim == 0)[0] + nleading
        else:
            gap_inds = np.where(allzero == 0)[0]
        mkrdata[marker + '_gaps'] = gap_inds
    return mkrdata


def _get_c3d_subject_param(acq, param):
    try:
        param = _get_c3d_metadata_field(acq, 'PROCESSING', param)[0]
    except ValueError:
        logger.warning('Cannot get subject parameter %s' % param)
        param = None
    return param


def get_metadata(c3dfile):
    """ Read trial and subject metadata """
    trialname = os.path.basename(os.path.splitext(c3dfile)[0])
    sessionpath = os.path.dirname(c3dfile)
    acq = _get_c3dacq(c3dfile)
    # frame offset (start of trial data in frames)
    offset = acq.GetFirstFrame()
    lastfr = acq.GetLastFrame()
    length = lastfr - offset + 1  # or acq.GetPointFrameNumber()
    framerate = acq.GetPointFrequency()
    analograte = acq.GetAnalogFrequency()
    samplesperframe = acq.GetNumberAnalogSamplePerFrame()

    # count forceplates
    fpe = btk.btkForcePlatformsExtractor()
    fpe.SetInput(acq)
    fpe.Update()
    n_forceplates = len(list(btk.Iterate(fpe.GetOutput())))

    # get markers
    markers = _get_c3d_metadata_field(acq, 'POINT', 'LABELS')
    # not sure what the '*xx' markers are, but delete them for now
    markers = [m for m in markers if m[0] != '*']

    #  get events
    rstrikes, lstrikes, rtoeoffs, ltoeoffs = [], [], [], []
    for i in btk.Iterate(acq.GetEvents()):
        if i.GetLabel() == "Foot Strike":
            if i.GetContext() == "Right":
                rstrikes.append(i.GetFrame())
            elif i.GetContext() == "Left":
                lstrikes.append(i.GetFrame())
            else:
                raise GaitDataError("Unknown context on foot strike event")
        elif i.GetLabel() == "Foot Off":
            if i.GetContext() == "Right":
                rtoeoffs.append(i.GetFrame())
            elif i.GetContext() == "Left":
                ltoeoffs.append(i.GetFrame())
            else:
                raise GaitDataError("Unknown context on foot strike event")

    # get subject info
    try:
        name = _get_c3d_metadata_field(acq, 'SUBJECTS', 'NAMES')[0]
    except ValueError:
        logger.warning('Cannot get subject name')
        name = u'Unknown'

    par_names = _get_c3d_metadata_subfields(acq, 'PROCESSING')
    subj_params = defaultdict(lambda: None)
    subj_params.update({par: _get_c3d_subject_param(acq, par) for
                        par in par_names})

    # sort events (may be in wrong temporal order, at least in c3d files)
    for li in [lstrikes, rstrikes, ltoeoffs, rtoeoffs]:
        li.sort()

    return {'trialname': trialname, 'sessionpath': sessionpath,
            'offset': offset, 'framerate': framerate, 'analograte': analograte,
            'name': name, 'subj_params': subj_params, 'lstrikes': lstrikes,
            'rstrikes': rstrikes, 'ltoeoffs': ltoeoffs, 'rtoeoffs': rtoeoffs,
            'length': length, 'samplesperframe': samplesperframe,
            'n_forceplates': n_forceplates, 'markers': markers}


def get_model_data(c3dfile, model):
    modeldata = dict()
    acq = _get_c3dacq(c3dfile)
    var_dims = (3, acq.GetPointFrameNumber())
    for var in model.read_vars:
        try:
            vals = acq.GetPoint(var).GetValues()
            modeldata[var] = np.transpose(np.squeeze(vals))
        except RuntimeError:
            if model.is_optional_var(var):
                logger.info('cannot read optional variable %s, returning nans'
                            % var)
                data = np.empty(var_dims)
                data[:] = np.nan
                modeldata[var] = data
            else:
                raise GaitDataError('Cannot find model variable %s in %s' %
                                    (var, c3dfile))
        # c3d stores scalars as last dim of 3-d array
        if model.read_strategy == 'last':
            modeldata[var] = modeldata[var][2, :]
    return modeldata


def get_forceplate_data(c3dfile):
    logger.debug('reading forceplate data from %s' % c3dfile)
    read_chs = ['Fx', 'Fy', 'Fz', 'Mx', 'My', 'Mz']
    acq = _get_c3dacq(c3dfile)
    fpe = btk.btkForcePlatformsExtractor()
    fpe.SetInput(acq)
    fpe.Update()
    fpdata = list()
    nplate = 0
    for plate in btk.Iterate(fpe.GetOutput()):
        logger.debug('reading from plate %d' % nplate)
        nplate += 1
        if plate.GetType() != 2:
            # Nexus should always write forceplates as type 2
            raise GaitDataError('Only type 2 forceplates are '
                                'supported for now')
        rawdata = dict()
        data = dict()
        for ch in btk.Iterate(plate.GetChannels()):
            label = ch.GetLabel()[-3:-1]  # strip descriptor and plate number
            rawdata[label] = np.squeeze(ch.GetData().GetValues())
        if not all([ch in rawdata for ch in read_chs]):
            raise GaitDataError('could not read force/moment data')
        F = np.stack([rawdata['Fx'], rawdata['Fy'], rawdata['Fz']], axis=1)
        M = np.stack([rawdata['Mx'], rawdata['My'], rawdata['Mz']], axis=1)
        # this should be the plate thickness (from moment origin to physical
        # origin) needed for center of pressure calculations
        dz = np.abs(plate.GetOrigin()[2])
        cop = center_of_pressure(F, M, dz)  # in plate local coords
        Ftot = np.linalg.norm(F, axis=1)
        # locations of +x+y, -x+y, -x-y, +x-y plate corners in world coords
        # (in that order)
        cor = plate.GetCorners()
        wT = np.mean(cor, axis=1)  # translation vector, plate -> world
        # upper and lower bounds of forceplate
        ub = np.max(cor, axis=1)
        lb = np.min(cor, axis=1)
        # plate unit vectors in world system
        px = cor[:, 0] - cor[:, 1]
        py = cor[:, 0] - cor[:, 3]
        pz = np.array([0, 0, -1])
        P = np.stack([px, py, pz], axis=1)
        wR = P / np.linalg.norm(P, axis=0)  # rotation matrix, plate -> world
        # check whether cop stays inside forceplate area and clip if necessary
        cop_w = change_coords(cop, wR, wT)
        cop_wx = np.clip(cop_w[:, 0], lb[0], ub[0])
        cop_wy = np.clip(cop_w[:, 1], lb[1], ub[1])
        if not (cop_wx == cop_w[:, 0]).all() and (cop_wy == cop_w[:, 1]).all():
            logger.warning('center of pressure outside forceplate '
                           'bounds, clipping to plate')
            cop[:, 0] = cop_wx
            cop[:, 1] = cop_wy
        # XXX moment and force transformations may still be wrong
        data['F'] = change_coords(-F, wR, 0)  # not sure why sign flip needed
        data['Ftot'] = Ftot
        data['M'] = change_coords(-M, wR, 0)  # not sure why sign flip needed
        data['CoP'] = cop_w
        data['upperbounds'] = ub
        data['lowerbounds'] = lb
        data['wR'] = wR
        data['wT'] = wT
        data['cor_full'] = cor.T
        fpdata.append(data)
    return fpdata



# -*- coding: utf-8 -*-
"""
Video file related code.

@author: Jussi (jnu@iki.fi)
"""

from builtins import str
from builtins import zip
import logging
import os.path as op
import glob
import itertools

from gaitutils import cfg

logger = logging.getLogger(__name__)


def get_trial_videos(trialfile, camera_label=None, vid_ext=None, overlay=None):
    """Return list of video files for trial file. File may be c3d or enf etc"""
    trialbase = op.splitext(trialfile)[0]
    # XXX: should really be case insensitive, but does not matter on Windows
    vid_exts = ['avi', 'ogv']
    globs_ = ('%s*%s' % (trialbase, vid_ext) for vid_ext in vid_exts)
    vids = itertools.chain.from_iterable(glob.iglob(glob_) for glob_ in globs_)
    if camera_label is not None:
        vids = _filter_by_label(vids, camera_label)
    if vid_ext is not None:
        vids = _filter_by_extension(vids, vid_ext)
    if overlay is not None:
        vids = _filter_by_overlay(vids, overlay)
    return list(vids)


def _filter_by_label(vids, camera_label):
    """Filter videos by camera label"""
    ids = [id for id, label in cfg.general.camera_labels.items() if
           camera_label == label]
    vid_its = itertools.tee(vids, len(ids))  # need to reuse vids iterator
    for id, vids_ in zip(ids, vid_its):
        for vid in _filter_by_id(vids_, id):
            yield vid


def _filter_by_extension(vids, ext):
    """Filter videos by filename extension. Case insensitive"""
    for vid in vids:
        if op.splitext(vid)[1].upper() == ext.upper():
            yield vid


def _filter_by_id(vids, camera_id):
    """Filter videos by id"""
    for vid in vids:
        if _camera_id(vid) == camera_id:
            yield vid


def _filter_by_overlay(vids, overlay=True):
    """Filter to get overlay or non-overlay videos"""
    for vid in vids:
        if 'overlay' in vid:
            if overlay:
                yield vid
        else:
            if not overlay:
                yield vid


def _camera_id(fname):
    """ Returns camera id for a video file """
    fn_split = op.split(fname)[-1].split('.')
    if len(fn_split) < 3:
        raise ValueError('Unexpected video file name %s' % fname)
    return fn_split[-3]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""

Script to create the full pdf gait report.
Note: specific to the Helsinki gait lab.


@author: Jussi (jnu@iki.fi)
"""
from __future__ import absolute_import

import logging
import argparse
import os.path as op
import matplotlib.pyplot as plt
from matplotlib.backends.backend_pdf import PdfPages
from collections import defaultdict

from gaitutils import (Plotter, cfg, register_gui_exception_handler, layouts,
                       numutils, normaldata, sessionutils, GaitDataError,
                       nexus)

from gaitutils.scripts import (nexus_kin_consistency,
                               nexus_emg_consistency,
                               nexus_musclelen_consistency,
                               nexus_kin_average,
                               nexus_trials_velocity,
                               nexus_time_distance_vars)


logger = logging.getLogger(__name__)

sort_field = 'NOTES'  # sort trials by the given Eclipse key
page_size = (11.69, 8.27)  # report page size
# create some separate PDFs for inclusion in Polygon report etc.
make_separate_pdfs = False


def _add_footer(fig, txt):
    fig.text(0, 0, txt, fontsize=8, color='black', ha='left', va='bottom')


def _add_header(fig, txt):
    fig.text(0, 1, txt, fontsize=8, color='black', ha='left', va='top')


def _savefig(pdf, fig, header=None, footer=None):
    """add header/footer into page and save as A4"""
    if fig is None:
        return
    if header is not None:
        _add_header(fig, header)
    if footer is not None:
        _add_footer(fig, footer)
    fig.set_size_inches(page_size[0], page_size[1])
    pdf.savefig(fig)


def do_plot(sessionpath, info=None, pages=None):
    """Create the pdf report and save in session directory"""

    if info is None:
        info = defaultdict(lambda: '')
    fullname = info['fullname'] or ''
    hetu = info['hetu'] or ''
    session_description = info['session_description'] or ''
    if pages is None:
        pages = defaultdict(lambda: True)  # default: do all plots
    elif not any(pages.values()):
        raise ValueError('No pages to print')

    tagged_figs = list()
    repr_figs = list()
    eclipse_tags = dict()
    do_emg_consistency = False

    session_root, sessiondir = op.split(sessionpath)
    patient_code = op.split(session_root)[1]
    pdfname = sessiondir + '.pdf'
    pdfpath = op.join(sessionpath, pdfname)

    tagged_trials = sessionutils.get_c3ds(sessionpath, tags=cfg.eclipse.tags,
                                          trial_type='dynamic')
    if not tagged_trials:
        raise GaitDataError('No tagged trials found in %s' % sessiondir)
    session_t = sessionutils.get_session_date(sessionpath)
    logger.debug('session timestamp: %s', session_t)
    age = numutils.age_from_hetu(hetu, session_t) if hetu else None

    # make header page
    # timestr = time.strftime('%d.%m.%Y')  # current time, not currently used
    fig_hdr = plt.figure()
    ax = plt.subplot(111)
    plt.axis('off')
    title_txt = 'HUS Liikelaboratorio\n'
    title_txt += u'Kävelyanalyysin tulokset\n'
    title_txt += '\n'
    title_txt += u'Nimi: %s\n' % fullname
    title_txt += u'Henkilötunnus: %s\n' % (hetu if hetu else 'ei tiedossa')
    title_txt += u'Ikä mittaushetkellä: %s\n' % ('%d vuotta' % age if age
                                                   else 'ei tiedossa')
    title_txt += u'Mittaus: %s\n' % sessiondir
    if session_description:
        title_txt += u'Kuvaus: %s\n' % session_description
    title_txt += u'Mittauksen pvm: %s\n' % session_t.strftime('%d.%m.%Y')
    title_txt += u'Liikelaboratorion potilaskoodi: %s\n' % patient_code
    ax.text(.5, .8, title_txt, ha='center', va='center', weight='bold',
            fontsize=14)

    header = u'Nimi: %s Henkilötunnus: %s' % (fullname, hetu)
    musclelen_ndata = normaldata.normaldata_age(age)
    footer_musclelen = (u' Normaalidata: %s' % musclelen_ndata if
                        musclelen_ndata else u'')

    pl = Plotter()

    for c3d in tagged_trials:

        pl.open_trial(c3d)
        representative = pl.trial.eclipse_tag in cfg.eclipse.repr_tags

        # representative single trial plots
        if representative:
            if pages['TimeDistRepresentative']:
                fig = nexus_time_distance_vars.do_single_trial_plot(c3d,
                                                                    show=False)
                repr_figs.append(fig)

        # try to figure out whether we have any valid EMG signals
        emg_active = any([pl.trial.emg.status_ok(ch) for ch in
                          cfg.emg.channel_labels])

        if emg_active:

            if pages['EMGCons']:
                do_emg_consistency = True

            if pages['KinEMGMarked']:
                logger.debug('creating representative kin-EMG plots')
                # FIXME: the plotter logic is a bit weird here - it works
                # but old axes get recreated
                for side in pl.trial.fp_events['valid']:
                    side_str = {'R': 'right', 'L': 'left'}[side]
                    pl.layout = (cfg.layouts.lb_kinetics_emg_r if side == 'R'
                                 else cfg.layouts.lb_kinetics_emg_l)

                    maintitle = ('Kinetics-EMG (%s) for %s' % (side_str, pl.title_with_eclipse_info()))
                    fig = pl.plot_trial(maintitle=maintitle, show=False)
                    tagged_figs.append(fig)
                    eclipse_tags[fig] = (pl.trial.eclipse_data[sort_field])

                    # save individual pdf
                    if representative and make_separate_pdfs:
                        pdf_name = 'kinetics_EMG_%s_%s.pdf' % (pl.trial.trialname,
                                                               side_str)
                        logger.debug('creating %s' % pdf_name)
                        pl.create_pdf(pdf_name=pdf_name)

            if pages['EMGMarked']:
                logger.debug('creating representative EMG plots')
                maintitle = pl.title_with_eclipse_info('EMG plot for')
                layout = cfg.layouts.std_emg
                pl.layout = layouts.rm_dead_channels(pl.trial.emg, layout)
                fig = pl.plot_trial(maintitle=maintitle, show=False)
                tagged_figs.append(fig)
                eclipse_tags[fig] = (pl.trial.eclipse_data[sort_field])

                #  save individual pdf
                if representative and make_separate_pdfs:
                    pdf_prefix = 'EMG_'
                    pl.create_pdf(pdf_prefix=pdf_prefix)

    tagged_figs.sort(key=lambda fig: eclipse_tags[fig])

    # trial velocity plot
    fig_vel = None
    if pages['TrialVelocity']:
        logger.debug('creating velocity plot')
        fig_vel = nexus_trials_velocity.do_plot(sessionpath, show=False,
                                                make_pdf=False)

    # time-distance average
    fig_timedist_avg = None
    if pages['TimeDistAverage']:
        logger.debug('creating time-distance plot')
        fig_timedist_avg = (nexus_time_distance_vars.
                            do_session_average_plot(sessionpath=sessionpath,
                                                    show=False,
                                                    make_pdf=False))
    # consistency plots
    fig_kin_cons = None
    if pages['KinCons']:
        logger.debug('creating kin consistency plot')
        fig_kin_cons = nexus_kin_consistency.do_plot(sessions=[sessionpath],
                                                     show=False,
                                                     make_pdf=make_separate_pdfs,
                                                     backend='matplotlib')
    fig_musclelen_cons = None
    if pages['MuscleLenCons']:
        logger.debug('creating muscle length consistency plot')
        fig_musclelen_cons = nexus_musclelen_consistency.do_plot(sessionpath=sessionpath,
                                                                 show=False,
                                                                 age=age,
                                                                 make_pdf=make_separate_pdfs)
    fig_emg_cons = None
    if do_emg_consistency:
        logger.debug('creating EMG consistency plot')        
        fig_emg_cons = nexus_emg_consistency.do_plot(sessionpath=sessionpath,
                                                     show=False,
                                                     make_pdf=make_separate_pdfs,
                                                     backend='matplotlib')

    # average plots
    figs_kin_avg = list()
    if pages['KinAverage']:
        figs_kin_avg = nexus_kin_average.do_plot(sessionpath=sessionpath,
                                                 show=False, make_pdf=False)

    logger.debug('creating multipage pdf %s' % pdfpath)
    with PdfPages(pdfpath) as pdf:
        _savefig(pdf, fig_hdr)
        _savefig(pdf, fig_vel, header)
        _savefig(pdf, fig_timedist_avg, header)
        _savefig(pdf, fig_kin_cons, header)
        _savefig(pdf, fig_musclelen_cons, header, footer_musclelen)
        _savefig(pdf, fig_emg_cons, header)
        for fig in figs_kin_avg:
            _savefig(pdf, fig, header)
        for fig in repr_figs:
            _savefig(pdf, fig, header)
        for fig in tagged_figs:
            _savefig(pdf, fig, header)

    # close all created figures, otherwise they'll pop up on next show() call
    plt.close('all')


if __name__ == '__main__':
    logging.basicConfig(level=logging.DEBUG)
    logging.getLogger('matplotlib.font_manager').setLevel(logging.WARNING)
    logging.getLogger('gaitutils.utils').setLevel(logging.WARNING)
    register_gui_exception_handler()

    parser = argparse.ArgumentParser()
    parser.add_argument('--session', type=str)
    args = parser.parse_args()

    session = args.session or nexus.get_sessionpath()
    do_plot(session)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Created on Tue Apr 28 11:37:51 2015

Plot currently loaded Nexus trial.

@author: Jussi (jnu@iki.fi)
"""

import logging
import argparse

from gaitutils import (Plotter, cfg, register_gui_exception_handler, layouts,
                       trial, plot_plotly, GaitDataError)


def do_plot(layout_name=None, backend=None, model_cycles=None, emg_cycles=None,
            maintitle=None, from_c3d=True):

    if layout_name is None:
        layout_name = 'lb_kinematics'

    try:
        layout = getattr(cfg.layouts, layout_name)
    except AttributeError:
        raise GaitDataError('No such layout %s' % layout_name)

    tr = trial.nexus_trial(from_c3d=from_c3d)
    model_cycles = ('unnormalized' if tr.is_static else
                    model_cycles)
    emg_cycles = ('unnormalized' if tr.is_static else
                  emg_cycles)

    # remove dead EMG channel
    # should be a no-op for non-EMG layouts, but restrict it to '*EMG*' anyway
    if 'EMG' in layout_name.upper():
        layout = layouts.rm_dead_channels(tr.emg, layout)

    if backend is None:
        backend = cfg.plot.backend

    if backend == 'matplotlib':
        pl = Plotter()
        pl.layout = layout
        pl.plot_trial(tr, model_cycles=model_cycles, emg_cycles=emg_cycles,
                      show=False)
        pl.show()

    elif backend == 'plotly':
        plot_plotly.plot_trials_browser([tr], layout, model_cycles=model_cycles,
                                        emg_cycles=emg_cycles,
                                        legend_type='short_name_with_cyclename')

    else:
        raise ValueError('Invalid plotting backend %s' % cfg.plot.backend)


if __name__ == '__main__':

    logging.basicConfig(level=logging.DEBUG)
    register_gui_exception_handler()
    parser = argparse.ArgumentParser()
    parser.add_argument('--layout', type=str)
    parser.add_argument('--backend', type=str)
    parser.add_argument('--unnorm', action='store_true')
    parser.add_argument('--model_cycles', type=str)
    parser.add_argument('--emg_cycles', type=str)
    args = parser.parse_args()

    if args.unnorm:
        args.model_cycles = args.emg_cycles = 'unnormalized'

    do_plot(layout_name=args.layout, backend=args.backend,
            model_cycles=args.model_cycles, emg_cycles=args.emg_cycles)
